"""
Annotation searcher for static comment searching via Stevedore plugins.
"""

import os

from code_annotations.base import BaseSearch


class StaticSearch(BaseSearch):
    """
    Handles static code searching for annotations.
    """

    def search_extension(self, ext, file_handle, file_extensions_map, filename_extension):
        """
        Execute a search on the given file using the given extension.

        Args:
            ext: Extension to execute the search on
            file_handle: An open file handle search
            file_extensions_map: Dict mapping of extension names to configured filename extensions
            filename_extension: The filename extension of the file being searched

        Returns:
            Tuple of (extension name, list of found annotation dicts)
        """
        # Only search this file if we are configured for its extension
        if filename_extension in file_extensions_map[ext.name]:
            # Reset the read handle to the beginning of the file in case another
            # extension moved it
            file_handle.seek(0)

            ext_results = ext.obj.search(file_handle)

            if ext_results:
                return ext.name, ext_results

        return ext.name, None

    def _search_one_file(self, full_name, known_extensions, file_extensions_map, all_results):
        """
        Perform an annotation search on a single file, using all extensions it is configured for.

        Args:
            full_name: Complete filename
            known_extensions: List of all file name extensions we are configured to work on
            file_extensions_map: Mapping of file name extensions to Stevedore extensions
            all_results: A dict of annotations returned from search()
        """
        filename_extension = os.path.splitext(full_name)[1][1:]

        if filename_extension not in known_extensions:
            self.echo.echo_vvv(
                f"{filename_extension} is not a known extension, skipping ({full_name})."
            )
            return

        self.echo.echo_vvv(full_name)

        # TODO: This should probably be a generator so we don't have to store all results in memory
        with open(full_name) as file_handle:
            # Call search_extension on all loaded extensions
            results = self.config.mgr.map(self.search_extension, file_handle, file_extensions_map, filename_extension)

            # Strip out plugin name, as it's already in the annotation
            results = [r for _, r in results]

            # Format and add the results to our running full set
            self.format_file_results(all_results, results)

    def search(self):
        """
        Walk the source tree, send known file types to extensions.

        Returns:
            Dict of all found annotations keyed by filename
        """
        # Index the results by extension name
        file_extensions_map = {}
        known_extensions = set()
        for extension_name in self.config.extensions:
            file_extensions_map[extension_name] = self.config.extensions[extension_name]
            known_extensions.update(self.config.extensions[extension_name])

        all_results = {}

        if os.path.isfile(self.config.source_path):
            self._search_one_file(self.config.source_path, known_extensions, file_extensions_map, all_results)
        else:
            for root, _, files in os.walk(self.config.source_path):
                for filename in files:
                    full_name = os.path.join(root, filename)
                    self._search_one_file(full_name, known_extensions, file_extensions_map, all_results)

        return all_results



"""
Stevedore extension for static annotation searching in Javascript files.
"""
import re

from code_annotations.extensions.base import SimpleRegexAnnotationExtension


class JavascriptAnnotationExtension(SimpleRegexAnnotationExtension):
    """
    Annotation extension for Javascript source files.
    """

    extension_name = 'javascript'

    lang_comment_definition = {
        'multi_start': re.escape('/*'),
        'multi_end': re.escape('*/'),
        'single': re.escape('//')
    }



"""
Click command to do static annotation searching via Stevedore plugins.
"""
import datetime
import errno
import os
import re
from abc import ABCMeta, abstractmethod

import yaml
from stevedore import named

from code_annotations import annotation_errors
from code_annotations.exceptions import ConfigurationException
from code_annotations.helpers import VerboseEcho


class AnnotationConfig:
    """
    Configuration shared among all Code Annotations commands.
    """

    def __init__(self, config_file_path, report_path_override=None, verbosity=1, source_path_override=None):
        """
         Initialize AnnotationConfig.

        Args:
            config_file_path: Path to the configuration file
            report_path_override: Path to write reports to, if overridden on the command line
            verbosity: Verbosity level from the command line
            source_path_override: Path to search if we're static code searching, if overridden on the command line
        """
        self.groups = {}
        self.choices = {}
        self.optional_groups = []
        self.annotation_tokens = []
        self.annotation_regexes = []
        self.mgr = None

        # Global logger, other objects can hold handles to this
        self.echo = VerboseEcho()

        with open(config_file_path) as config_file:
            raw_config = yaml.safe_load(config_file)

        self._check_raw_config_keys(raw_config)

        self.safelist_path = raw_config['safelist_path']
        self.extensions = raw_config['extensions']

        self.verbosity = verbosity
        self.echo.set_verbosity(verbosity)

        self.report_path = report_path_override if report_path_override else raw_config['report_path']
        self.echo(f"Configured for report path: {self.report_path}")

        self.source_path = source_path_override if source_path_override else raw_config['source_path']
        self.echo(f"Configured for source path: {self.source_path}")

        self._configure_coverage(raw_config.get('coverage_target', None))
        self.report_template_dir = raw_config.get('report_template_dir')
        self.rendered_report_dir = raw_config.get('rendered_report_dir')
        self.rendered_report_file_extension = raw_config.get('rendered_report_file_extension')
        self.rendered_report_source_link_prefix = raw_config.get('rendered_report_source_link_prefix')

        self._configure_annotations(raw_config)
        self._configure_extensions()

    def _check_raw_config_keys(self, raw_config):
        """
        Validate that all required keys exist in the configuration file.

        Args:
            raw_config: Python representation of the YAML config file

        Raises:
            ConfigurationException on any missing keys
        """
        errors = []
        for k in ('report_path', 'source_path', 'safelist_path', 'annotations', 'extensions'):
            if k not in raw_config:
                errors.append(k)

        if errors:
            raise ConfigurationException(
                'The following required keys are missing from the configuration file: \n{}'.format(
                    '\n'.join(errors)
                )
            )

    def _is_annotation_group(self, token_or_group):
        """
        Determine if an annotation is a group or not.

        Args:
            token_or_group: The annotation being checked

        Returns:
            True if the type of the annotation is correct for a group, otherwise False
        """
        return isinstance(token_or_group, list)

    def _is_choice_group(self, token_or_group):
        """
        Determine if an annotation is a choice group.

        Args:
            token_or_group: The annotation being checked

        Returns:
            True if the type of the annotation is correct for a choice group, otherwise False
        """
        return isinstance(token_or_group, dict) and "choices" in token_or_group

    def _is_optional_group(self, token_or_group):
        """
        Determine if an annotation is an optional group.

        Args:
            token_or_group: The annotation being checked

        Returns:
            True if the annotation is optional, otherwise False.
        """
        return isinstance(token_or_group, dict) and bool(token_or_group.get("optional"))

    def _is_annotation_token(self, token_or_group):
        """
        Determine if an annotation has the right format.

        Args:
            token_or_group: The annotation being checked

        Returns:
            True if the type of the annotation is correct for a text type, otherwise False
        """
        if token_or_group is None:
            return True
        if isinstance(token_or_group, dict):
            # If annotation is a dict, only a few keys are tolerated
            return set(token_or_group.keys()).issubset({"choices", "optional"})
        return False

    def _add_annotation_token(self, token):
        if token in self.annotation_tokens:
            raise ConfigurationException(f'{token} is configured more than once, tokens must be unique.')
        self.annotation_tokens.append(token)

    def _configure_coverage(self, coverage_target):
        """
        Set coverage_target to the specified value.

        Args:
            coverage_target:

        Returns:

        """
        if coverage_target:
            try:
                self.coverage_target = float(coverage_target)
            except (TypeError, ValueError) as error:
                raise ConfigurationException(
                    f'Coverage target must be a number between 0 and 100 not "{coverage_target}".'
                ) from error

            if self.coverage_target < 0.0 or self.coverage_target > 100.0:
                raise ConfigurationException(
                    f'Invalid coverage target. {self.coverage_target} is not between 0 and 100.'
                )
        else:
            self.coverage_target = None

    def _configure_group(self, group_name, group):
        """
        Perform group configuration and add annotations from the group to global configuration.

        Args:
            group_name: The name of the group (the key in the configuration dictionary)
            group: The list of annotations that comprise the group

        Raises:
            TypeError if the group is misconfigured
        """
        self.groups[group_name] = []

        if not group or len(group) == 1:
            raise ConfigurationException(f'Group "{group_name}" must have more than one annotation.')

        for annotation in group:
            for annotation_token in annotation:
                annotation_value = annotation[annotation_token]

                # Otherwise it should be a text type, if not then error out
                if not self._is_annotation_token(annotation_value):
                    raise ConfigurationException(f'{annotation} is an unknown annotation type.')
                # The annotation comment is a choice group
                if self._is_choice_group(annotation_value):
                    self._configure_choices(annotation_token, annotation_value)
                # The annotation comment is not mandatory
                if self._is_optional_group(annotation_value):
                    self.optional_groups.append(annotation_token)

                self.groups[group_name].append(annotation_token)
                self._add_annotation_token(annotation_token)
                self.annotation_regexes.append(re.escape(annotation_token))

    def _configure_choices(self, annotation_token, annotation):
        """
        Configure the choices list for an annotation.

        Args:
            annotation_token: The annotation token we are setting choices for
            annotation: The annotation body (list of choices)
        """
        self.choices[annotation_token] = annotation['choices']

    def _configure_annotations(self, raw_config):
        """
        Transform the configured annotations into more usable pieces and validate.

        Args:
            raw_config: The dictionary form of our configuration file
        Raises:
            TypeError if annotations are misconfigured
        """
        annotation_tokens = raw_config['annotations']

        for annotation_token_or_group_name in annotation_tokens:
            annotation = annotation_tokens[annotation_token_or_group_name]

            if self._is_annotation_group(annotation):
                self._configure_group(annotation_token_or_group_name, annotation)

            elif self._is_choice_group(annotation):
                self._configure_choices(annotation_token_or_group_name, annotation)
                self._add_annotation_token(annotation_token_or_group_name)
                self.annotation_regexes.append(re.escape(annotation_token_or_group_name))

            elif not self._is_annotation_token(annotation):  # pragma: no cover
                raise TypeError(
                    f'{annotation_token_or_group_name} is an unknown type, must be strings or lists.'
                )
            else:
                self._add_annotation_token(annotation_token_or_group_name)
                self.annotation_regexes.append(re.escape(annotation_token_or_group_name))

        self.echo.echo_v(f"Groups configured: {self.groups}")
        self.echo.echo_v(f"Choices configured: {self.choices}")
        self.echo.echo_v(f"Annotation tokens configured: {self.annotation_tokens}")

    def _plugin_load_failed_handler(self, *args, **kwargs):
        """
        Handle failures to load an extension.

        Dumps the error and raises an exception. By default these
        errors just fail silently.

        Args:
            *args:
            **kwargs:

        Raises:
            ConfigurationException
        """
        self.echo(str(args), fg='red')
        self.echo(str(kwargs), fg='red')
        raise ConfigurationException('Failed to load a plugin, aborting.')

    def _configure_extensions(self):
        """
        Configure the Stevedore NamedExtensionManager.

        Raises:
            ConfigurationException
        """
        # These are the names of all of our configured extensions
        configured_extension_names = self.extensions.keys()

        # Load Stevedore extensions that we are configured for (and only those)
        self.mgr = named.NamedExtensionManager(
            names=configured_extension_names,
            namespace='annotation_finder.searchers',
            invoke_on_load=True,
            on_load_failure_callback=self._plugin_load_failed_handler,
            invoke_args=(self, self.echo),
        )

        # Output extension names listed in configuration
        self.echo.echo_vv("Configured extension names: {}".format(" ".join(configured_extension_names)))

        # Output found extension entry points from setup.py|cfg (whether or not they were loaded)
        self.echo.echo_vv("Stevedore entry points found: {}".format(str(self.mgr.list_entry_points())))

        # Output extensions that were actually able to load
        self.echo.echo_v("Loaded extensions: {}".format(" ".join([x.name for x in self.mgr.extensions])))

        if len(self.mgr.extensions) != len(configured_extension_names):
            raise ConfigurationException('Not all configured extensions could be loaded! Asked for {} got {}.'.format(
                configured_extension_names, self.mgr.extensions
            ))


class BaseSearch(metaclass=ABCMeta):
    """
    Base class for searchers.
    """

    def __init__(self, config):
        """
        Initialize for StaticSearch.

        Args:
            config: Configuration object
        """
        self.config = config
        self.echo = self.config.echo
        # errors contains formatted error messages
        self.errors = []
        # annotation_errors contains (annotation, AnnotationErrorType, args) tuples
        # This attribute may be parsed by 3rd-parties, such as edx-lint.
        self.annotation_errors = []

    def format_file_results(self, all_results, results):
        """
        Add all extensions' search results for a file to the overall results.

        Args:
            all_results: Aggregated results to add the results to
            results: Results of search() on a single file

        Returns:
            None, modifies all_results
        """
        for annotations in results:
            if not annotations:
                continue

            # TODO: The file_path should be the same for all of these results
            # so we should be able to optimize getting file_path and making
            # sure it exists in the dict to do this less often.
            file_path = annotations[0]['filename']

            if file_path not in all_results:  # pragma: no cover
                all_results[file_path] = []

            for annotation in annotations:
                # If this is a "choices" type of annotation, split the comment into a list.
                # Actually checking the choice validity happens later in _check_results_choices.
                if annotation['annotation_token'] in self.config.choices:
                    annotation['annotation_data'] = re.split(r',\s?|\s', annotation['annotation_data'])

            # TODO: De-dupe results? Should only be necessary if more than one
            # Stevedore extension is working on the same file type
            all_results[file_path].extend(annotations)

    def _check_results_choices(self, annotation):
        """
        Check that a search result has appropriate choices.

        If the following errors are found:
            - no choices
            - multiple of the same choice
            - a choice which is not configured

        This function will add the error to self.errors.

        Args:
            annotation: A single search result dict.
        """
        # Not a choice type of annotation, nothing to do
        if annotation['annotation_token'] not in self.config.choices:
            return None

        token = annotation['annotation_token']
        found_valid_choices = []

        # If the line begins with an annotation token that should have choices, but has no text after the token,
        # the first split will be empty.
        if annotation['annotation_data'][0] != "":
            for choice in annotation['annotation_data']:
                if choice not in self.config.choices[token]:
                    self._add_annotation_error(
                        annotation,
                        annotation_errors.InvalidChoice,
                        (choice, token, self.config.choices[token])
                    )
                elif choice in found_valid_choices:
                    self._add_annotation_error(annotation, annotation_errors.DuplicateChoiceValue, (choice,))
                else:
                    found_valid_choices.append(choice)
        else:
            self._add_annotation_error(
                annotation,
                annotation_errors.MissingChoiceValue,
                (token, self.config.choices[token])
            )
        return None

    def _get_group_children(self):
        """
        Create a list of all annotation tokens that are part of a group.

        Returns:
            List of annotation tokens that are configured to be in groups
        """
        group_children = []

        for group in self.config.groups:
            group_children.extend(self.config.groups[group])

        return group_children

    def _get_group_for_token(self, token):
        """
        Find out which group, if any, an annotation token belongs to.

        Args:
            token: Annotation token to search for

        Returns:
            The group name, or None if it doesn't belong to a group.
        """
        for group in self.config.groups:
            if token in self.config.groups[group]:
                return group
        return None

    def check_results(self, all_results):
        """
        Spin through all search results, confirm that they all match configuration.

        If errors are found they are added to self.errors.

        Args:
            all_results: Dict of annotations found in search()

        Returns:
            Boolean indicating whether or not any errors were found
        """
        self.echo.pprint(all_results, indent=3, verbosity_level=2)

        # Spin through the search results
        for filename in all_results:
            for annotations in self.iter_groups(all_results[filename]):
                self.check_group(annotations)
        return not self.errors

    def iter_groups(self, annotations):
        """
        Iterate on groups of annotations.

        Annotations are considered as a group when they all have the same `line_number` and optional
        `extra['object_id']`. The line number points to the beginning of the annotation group. The `object_id` is set
        mostly for annotations parsed from a safelist.

        Yield:
            annotations (annotation list)
        """
        current_group = []
        current_line_number = None
        current_object_id = None
        for annotation in annotations:
            line_number = annotation["line_number"]
            object_id = annotation.get("extra", {}).get("object_id")
            line_number_changed = line_number != current_line_number
            object_id_changed = object_id != current_object_id
            if line_number_changed or object_id_changed:
                if current_group:
                    yield current_group
                current_group.clear()
            current_group.append(annotation)
            current_line_number = line_number
            current_object_id = object_id

        if current_group:
            yield current_group

    def check_group(self, annotations):
        """
        Perform several linting checks on a group of annotations.

        The following checks are performed:

        - Choice fields should have a valid value
        - Annotation tokens are valid
        - There is no duplicate
        - All non-optional tokens are present
        """
        found_tokens = set()
        group_tokens = []
        group_name = None
        for annotation in annotations:
            token = annotation["annotation_token"]
            if not group_name:
                group_name = self._get_group_for_token(token)
                if group_name:
                    group_tokens = self.config.groups[group_name]

            # Check if choice field
            self._check_results_choices(annotation)

            # Check token belongs to group
            if group_name:
                if token not in group_tokens:
                    self._add_annotation_error(
                        annotation,
                        annotation_errors.InvalidToken,
                        (
                           token,
                           group_name,
                           group_tokens
                        )
                    )

            # Check for duplicates
            if token in found_tokens:
                self._add_annotation_error(
                    annotation,
                    annotation_errors.DuplicateToken,
                    (token,)
                )
            if group_name:
                found_tokens.add(token)

        # Check for missing tokens
        for token in group_tokens:
            if token not in self.config.optional_groups:
                if token not in found_tokens:
                    self._add_annotation_error(
                        annotations[0],
                        annotation_errors.MissingToken,
                        (token,)
                    )

    def _add_annotation_error(self, annotation, error_type, args=None):
        """
        Add an error message to self.errors, formatted nicely.

        Args:
            annotation: A single annotation dict found in search()
            error_type (annotation_errors.AnnotationErrorType): error type from which the error message will be
                generated.
            args (tuple): arguments for error message formatting.
        """
        args = args or tuple()
        error_message = error_type.message % args
        location = annotation.get("extra", {}).get("object_id", annotation["line_number"])
        message = "{}::{}: {}".format(annotation['filename'], location, error_message)
        self.annotation_errors.append((annotation, error_type, args))
        self._add_error(message)

    def _add_error(self, message):
        """
        Add an error message to self.errors.

        Args:
            message: Custom error message to be added
        """
        self.errors.append(message)

    @abstractmethod
    def search(self):
        """
        Walk the source tree, send known file types to extensions.

        Returns:
            Dict of {filename: annotations} for all files with found annotations.
        """

    def _format_results_for_report(self, all_results):
        """
        Format the given results dict for reporting purposes.

        Args:
            all_results: Dict of all results found in a search

        Returns:
            Dict of results arranged for reporting
        """
        formatted_results = {}
        current_group_id = 0
        for filename in all_results:
            self.echo.echo_vv(f"report_format: formatting {filename}")
            formatted_results[filename] = []
            for annotation_group in self.iter_groups(all_results[filename]):
                current_group_id += 1
                for annotation_index, annotation in enumerate(annotation_group):
                    token = annotation['annotation_token']
                    self.echo.echo_vvv(f"report_format: formatting annotation token {token}")
                    if annotation_index == 0:
                        self.echo.echo_vv('Starting group id {} for "{}" token "{}", line {}'.format(
                            current_group_id, annotation_group, token, annotation["line_number"])
                        )
                    self.echo.echo_vv(
                        "report_format: Adding {} to group id {}".format(
                            token,
                            current_group_id
                        )
                    )
                    annotation["report_group_id"] = current_group_id
                    formatted_results[filename].append(annotation)

        return formatted_results

    def report(self, all_results, report_prefix=''):
        """
        Generate the YAML report of all search results.

        Args:
            all_results: Dict of found annotations, indexed by filename
            report_prefix: Prefix to add to report filename

        Returns:
            Filename of generated report
        """
        self.echo.echo_vv(yaml.dump(all_results, default_flow_style=False))

        now = datetime.datetime.now()
        report_filename = os.path.join(self.config.report_path, '{}{}.yaml'.format(
            report_prefix, now.strftime('%Y-%d-%m-%H-%M-%S')
        ))

        formatted_results = self._format_results_for_report(all_results)

        self.echo(f"Generating report to {report_filename}")

        try:
            os.makedirs(self.config.report_path)
        except OSError as e:  # pragma: no cover
            if e.errno != errno.EEXIST:
                raise

        with open(report_filename, 'w+') as report_file:
            yaml.safe_dump(formatted_results, report_file, default_flow_style=False)

        return report_filename



"""
Abstract and base classes to support plugins.
"""
import re
from abc import ABCMeta, abstractmethod

from code_annotations.helpers import clean_abs_path, clean_annotation, get_annotation_regex


class AnnotationExtension(metaclass=ABCMeta):
    """
    Abstract base class that annotation extensions will inherit from.
    """

    extension_name = None

    def __init__(self, config, echo):
        """
        Initialize this base object, save a handle to configuration.

        Args:
            config: The configuration object
            echo: VerboseEcho object used for logging
        """
        self.config = config
        self.ECHO = echo

    @abstractmethod
    def search(self, file_handle):  # pragma: no cover
        """
        Search for annotations in the given file.
        """
        raise NotImplementedError('search called on base class!')


class SimpleRegexAnnotationExtension(AnnotationExtension, metaclass=ABCMeta):
    """
    Abstract base class for languages that have comments which can be reasonably searched using regular expressions.
    """

    # These are the language-specific comment definitions that are defined in the child classes. See the
    # Javascript and Python extensions for examples.
    lang_comment_definition = None

    # This format string/regex finds all comments in the file. The format tokens will be replaced with the
    # language-specific comment definitions defined in the sub-classes.
    #
    # Match groupdict will contain two named subgroups: 'comment' and 'prefixed_comment', of which at most
    # one will be non-None.
    comment_regex_fmt = r"""
        {multi_start}           # start of the language-specific multi-line comment (ex. /*)
        (?P<comment>            # Look for a multiline comment
            [\d\D]*?            # capture all of the characters...
        )
        {multi_end}             # until you find the end of the language-specific multi-line comment (ex. */)
        |                       # If you don't find any of those...
        (?P<prefixed_comment>   # Look for a group of single-line comments
            (?:                 # Non-capture mode
                {single}        # start by finding the single-line comment token (ex. //)
                .*              # and capture all characters until the end of the line
                \n?             # followed by an optional carriage return
                \ *             # and some empty space
            )*                  # multiple times
        )
    """

    def __init__(self, config, echo):
        """
        Set up the extension and create the regexes used to do searches.

        Args:
            config: The configuration dict
            echo: VerboseEcho object for logging
        """
        super().__init__(config, echo)

        if self.lang_comment_definition is None:  # pragma: no cover
            raise ValueError('Subclasses of SimpleRegexAnnotationExtension must define lang_comment_definition!')

        self.comment_regex = re.compile(
            self.comment_regex_fmt.format(**self.lang_comment_definition),
            flags=re.VERBOSE
        )
        self.prefixed_comment_regex = re.compile(
            r"^ *{single}".format(**self.lang_comment_definition),
            flags=re.MULTILINE
        )

        # Parent class will allow this class to populate self.strings_to_search via
        # calls to _add_annotation_token or _add_annotation_group for each configured
        # annotation.
        self.query = get_annotation_regex(self.config.annotation_regexes)

        self.ECHO.echo_v(f"{self.extension_name} extension regex query: {self.query.pattern}")

    def search(self, file_handle):
        """
        Search for annotations in the given file.

        Args:
            file_handle: Handle for the file to validate

        Returns:
            List of dicts describing found annotations, or an empty list
        """
        txt = file_handle.read()

        found_annotations = []

        # Fast out if no annotations exist in the file
        if any(anno in txt for anno in self.config.annotation_tokens):
            fname = clean_abs_path(file_handle.name, self.config.source_path)

            # Iterate on all comments: both prefixed- and non-prefixed.
            for match in self.comment_regex.finditer(txt):
                # Get the line number by counting newlines + 1 (for the first line).
                # Note that this is the line number of the beginning of the comment, not the
                # annotation token itself.
                line = txt.count('\n', 0, match.start()) + 1

                comment_content = self._find_comment_content(match)
                for inner_match in self.query.finditer(comment_content):
                    try:
                        annotation_token = inner_match.group('token')
                        annotation_data = inner_match.group('data')
                    except IndexError as error:
                        # pragma: no cover
                        raise ValueError('{}::{}: Could not find "data" or "token" groups. Found: {}'.format(
                            fname,
                            line,
                            inner_match.groupdict()
                        )) from error
                    annotation_token, annotation_data = clean_annotation(annotation_token, annotation_data)
                    found_annotations.append({
                        'found_by': self.extension_name,
                        'filename': fname,
                        'line_number': line,
                        'annotation_token': annotation_token,
                        'annotation_data': annotation_data,
                    })

        return found_annotations

    def _find_comment_content(self, match):
        """
        Return the comment content as text.

        Args:
            match (sre.SRE_MATCH): one of the matches of the self.comment_regex regular expression.
        """
        comment_content = match.groupdict()["comment"]
        if comment_content:
            return comment_content

        # Find single-line comments and strip comment tokens
        comment_content = match.groupdict()["prefixed_comment"]
        return self._strip_single_line_comment_tokens(comment_content)

    def _strip_single_line_comment_tokens(self, content):
        """
        Strip the leading single-line comment tokens from a comment text.

        Args:
            content (str): token-prefixed multi-line comment string.
        """
        return self.prefixed_comment_regex.sub("", content)



"""
Stevedore extension for static annotation searching in Python files.
"""
import re

from code_annotations.extensions.base import SimpleRegexAnnotationExtension


class PythonAnnotationExtension(SimpleRegexAnnotationExtension):
    """
    Annotation extension for Python source files.
    """

    extension_name = 'python'

    lang_comment_definition = {
        'multi_start': re.escape('"""'),
        'multi_end': re.escape('"""'),
        'single': re.escape('#')
    }



"""
Annotation searcher for Django model comment searching Django introspection.
"""
import inspect
import os
import sys

import django
import yaml
from django.apps import apps
from django.db import models

from code_annotations.base import BaseSearch
from code_annotations.helpers import clean_annotation, fail, get_annotation_regex

DEFAULT_SAFELIST_FILE_PATH = '.annotation_safe_list.yml'


class DjangoSearch(BaseSearch):
    """
    Handles Django model comment searching for annotations.
    """

    def __init__(self, config):
        """
        Initialize for DjangoSearch.

        Args:
            config: Configuration file path
        """
        super().__init__(config)
        self.local_models, self.non_local_models, total, needing_annotation = self.get_models_requiring_annotations()
        self.model_counts = {
            'total': total,
            'annotated': 0,
            'unannotated': 0,
            'needing_annotation': len(needing_annotation),
            'not_needing_annotation': total - len(needing_annotation),
            'safelisted': 0
        }
        self.uncovered_model_ids = set()
        self.echo.echo_vvv('Local models:\n   ' + '\n   '.join([str(m) for m in self.local_models]) + '\n')
        self.echo.echo_vvv('Non-local models:\n   ' + '\n   '.join([str(m) for m in self.non_local_models]) + '\n')
        self.echo.echo_vv('The following models require annotations:\n   ' + '\n   '.join(needing_annotation) + '\n')

    def _increment_count(self, count_type, incr_by=1):
        self.model_counts[count_type] += incr_by

    def seed_safelist(self):
        """
        Seed a new safelist file with all non-local models that need to be vetted.
        """
        if os.path.exists(self.config.safelist_path):
            fail(f'{self.config.safelist_path} already exists, not overwriting.')

        self.echo(
            'Found {} non-local models requiring annotations. Adding them to safelist.'.format(
                len(self.non_local_models))
        )

        safelist_data = {self.get_model_id(model): {} for model in self.non_local_models}

        with open(self.config.safelist_path, 'w') as safelist_file:
            safelist_comment = """
# This is a Code Annotations automatically-generated Django model safelist file.
# These models must be annotated as follows in order to be counted in the coverage report.
# See https://code-annotations.readthedocs.io/en/latest/safelist.html for more information.
#
# fake_app_1.FakeModelName:
#    ".. no_pii:": "This model has no PII"
# fake_app_2.FakeModel2:
#    ".. choice_annotation:": foo, bar, baz

"""
            safelist_file.write(safelist_comment.lstrip())
            yaml.safe_dump(safelist_data, stream=safelist_file, default_flow_style=False)

        self.echo(f'Successfully created safelist file "{self.config.safelist_path}".', fg='red')
        self.echo('Now, you need to:', fg='red')
        self.echo('  1) Make sure that any un-annotated models in the safelist are annotated, and', fg='red')
        self.echo('  2) Annotate any LOCAL models (see --list_local_models).', fg='red')

    def list_local_models(self):
        """
        Dump a list of models in the local code tree that need annotations to stdout.
        """
        if self.local_models:
            self.echo(
                'Listing {} local models requiring annotations:'.format(len(self.local_models))
            )
            self.echo.pprint(sorted([self.get_model_id(model) for model in self.local_models]), indent=4)
        else:
            self.echo('No local models requiring annotations.')

    def _append_model_annotations(self, model_type, model_id, query, model_annotations):
        """
        Find the given model's annotations in the file and add them to model_annotations.

        Args:
            model_type: The type of model we're searching
            model_id: The text representation of the model name from get_model_id
            query: The regex to run to find annotations in the docstring
            model_annotations: The running list of found annotations in search() that we are to add to
        """
        # Read in the source file to get the line number
        filename = inspect.getsourcefile(model_type)
        with open(filename) as file_handle:
            txt = file_handle.read()

        # Get the line number by counting newlines + 1 (for the first line).
        # Note that this is the line number of the beginning of the comment, not the
        # annotation token itself. We find based on the entire code content of the model
        # as that seems to be the only way to be sure we're getting the correct line number.
        # It is slow and should be replaced if we can find a better way that is accurate.
        line = txt.count('\n', 0, txt.find(inspect.getsource(model_type))) + 1

        for inner_match in query.finditer(model_type.__doc__):
            try:
                annotation_token = inner_match.group('token')
                annotation_data = inner_match.group('data')
            except IndexError as error:
                # pragma: no cover
                raise ValueError('{}: Could not find "data" or "token" groups. Found: {}'.format(
                    self.get_model_id(model_type),
                    inner_match.groupdict()
                )) from error
            annotation_token, annotation_data = clean_annotation(annotation_token, annotation_data)
            model_annotations.append({
                'found_by': "django",
                'filename': filename,
                'line_number': line,
                'annotation_token': annotation_token,
                'annotation_data': annotation_data,
                'extra': {
                    'object_id': model_id,
                    'full_comment': model_type.__doc__.strip()
                }
            })

    def _append_safelisted_model_annotations(self, safelisted_models, model_id, model_annotations):
        """
        Append the safelisted annotations for the given model id to model_annotations.

        Args:
            safelisted_models: The dict of models and their annotations loaded from the safelist
            model_id: The text representation of the model name from get_model_id
            model_annotations: The running list of found annotations in search() that we are to add to
        """
        for annotation in safelisted_models[model_id]:
            comment = safelisted_models[model_id][annotation]
            model_annotations.append({
                'found_by': "safelist",
                'filename': self.config.safelist_path,
                'line_number': 0,
                'annotation_token': annotation.strip(),
                'annotation_data': comment.strip(),
                'extra': {
                    'object_id': model_id,
                    'full_comment': str(safelisted_models[model_id])
                }
            })

    def _read_safelist(self):
        """
        Read the safelist and return the found models and their annotations.

        Returns:
            The Python representation of the safelist
        """
        if os.path.exists(self.config.safelist_path):
            self.echo(f'Found safelist at {self.config.safelist_path}. Reading.\n')
            with open(self.config.safelist_path) as safelist_file:
                safelisted_models = yaml.safe_load(safelist_file)
            self._increment_count('safelisted', len(safelisted_models))

            if safelisted_models:
                self.echo.echo_vv('   Safelisted models:\n   ' + '\n   '.join(safelisted_models))
            else:
                self.echo.echo_vv('   No safelisted models found.\n')

            return safelisted_models
        else:
            raise Exception('Safelist not found! Generate one with the --seed_safelist command.')

    def search(self):
        """
        Introspect the configured Django model docstrings for annotations.

        Returns:
            Dict of all found annotations keyed by filename
        """
        safelisted_models = self._read_safelist()
        annotation_tokens = self.config.annotation_tokens
        annotation_regexes = self.config.annotation_regexes
        query = get_annotation_regex(annotation_regexes)

        annotated_models = {}

        self.echo.echo_vv('Searching models and their parent classes...')

        # Walk all models and their parents looking for annotations
        for model in self.local_models.union(self.non_local_models):
            model_id = self.get_model_id(model)
            self.echo.echo_vv('   ' + model_id)
            hierarchy = inspect.getmro(model)
            model_annotations = []

            # If any annotations exist in the docstring add them to annotated_models
            for obj in hierarchy:
                if obj.__doc__ is not None:
                    if any(anno in obj.__doc__ for anno in annotation_tokens):
                        self.echo.echo_vvv('      ' + DjangoSearch.get_model_id(obj) + ' has annotations.')
                        self._append_model_annotations(obj, model_id, query, model_annotations)
                    else:
                        # Don't use get_model_id here, as this could be a base class below Model
                        self.echo.echo_vvv('      ' + str(obj) + ' has no annotations.')

            # If there are any annotations in the model, format them
            if model_annotations:
                self.echo.echo_vv("      {} has {} total annotations".format(model_id, len(model_annotations)))
                self._increment_count('annotated')
                if model_id in safelisted_models:
                    self._add_error(f"{model_id} is annotated, but also in the safelist.")
                self.format_file_results(annotated_models, [model_annotations])

            # The model is not in the safelist and is not annotated
            elif model_id not in safelisted_models:
                self._increment_count('unannotated')
                self.uncovered_model_ids.add(model_id)
                self.echo.echo_vv(f"      {model_id} has no annotations")

            # Otherwise it is not annotated and in the safelist
            else:
                if not safelisted_models[model_id]:
                    self.uncovered_model_ids.add(model_id)
                    self.echo.echo_vv(f"      {model_id} is in the safelist.")
                    self._add_error(f"{model_id} is in the safelist but has no annotations!")
                else:
                    self._increment_count('annotated')

                self._append_safelisted_model_annotations(safelisted_models, model_id, model_annotations)
                self.format_file_results(annotated_models, [model_annotations])

        return annotated_models

    def check_coverage(self):
        """
        Perform checking of coverage percentage based on stats collected in setup and search.

        Returns:
            Bool indicating whether or not the number of annotated models covers a percentage
                of total models needing annotations greater than or equal to the configured
                coverage_target.
        """
        self.echo("\nModel coverage report")
        self.echo("-" * 40)
        self.echo("Found {total} total models.".format(**self.model_counts))
        self.echo("{needing_annotation} needed annotation, {annotated} were annotated.".format(**self.model_counts))

        if self.model_counts['needing_annotation'] > 0:
            pct = float(self.model_counts['annotated']) / float(self.model_counts['needing_annotation']) * 100.0
            pct = round(pct, 1)
        else:
            pct = 100.0

        self.echo(f"Coverage is {pct}%\n")

        if self.uncovered_model_ids:
            displayed_uncovereds = list(self.uncovered_model_ids)
            displayed_uncovereds.sort()
            self.echo(
                "Coverage found {} uncovered models:\n   ".format(len(self.uncovered_model_ids)) +
                "\n   ".join(displayed_uncovereds)
            )

        if pct < float(self.config.coverage_target):
            self.echo(
                "\nCoverage threshold not met! Needed {}, actually {}!".format(
                    self.config.coverage_target,
                    pct
                )
            )
            return False

        return True

    @staticmethod
    def requires_annotations(model):
        """
        Return true if the given model actually requires annotations, according to PLAT-2344.
        """
        # Anything inheriting from django.models.Model will have a ._meta attribute. Our tests
        # inherit from object, which doesn't have it, and will fail below. This is a quick way
        # to early out on both.
        if not hasattr(model, '_meta'):
            return False

        return issubclass(model, models.Model) \
            and not (model is models.Model) \
            and not model._meta.abstract \
            and not model._meta.proxy

    @staticmethod
    def is_non_local(model):
        """
        Determine if the given model is non-local to the current IDA.

        Non-local models are all installed models which are not "local", by
        definition.  "Local" models are installed models which are defined in code
        which physically lives in the current codebase, where "current codebase"
        ostensibly refers to the code providing the currently active django
        project.

        Args:
            model (django.db.models.Model): A model to check.

        Returns:
            bool: True if the given model is non-local.
        """
        # If the model _was_ local to the current IDA repository, it should be
        # defined somewhere under sys.prefix + '/src/' or in a path that points to
        # the current checked-out code.  On Posix systems according to our testing,
        # non-local packages get installed to paths containing either
        # "site-packages" or "dist-packages".
        non_local_path_prefixes = []
        for path in sys.path:
            if 'dist-packages' in path or 'site-packages' in path:
                non_local_path_prefixes.append(path)
        model_source_path = inspect.getsourcefile(model)
        return model_source_path.startswith(tuple(non_local_path_prefixes))

    @staticmethod
    def get_model_id(model):
        """
        Construct the django standard model identifier in "app_label.ModelClassName" notation.

        Args:
            model (django.db.models.Model): A model for which to create an identifier.

        Returns:
            str: identifier string for the given model.
        """
        return f'{model._meta.app_label}.{model._meta.object_name}'

    @staticmethod
    def setup_django():
        """
        Prepare to make django library function calls.

        On behalf of the current Django project in the current working directory,
        setup/load the django framework, specified settings, and apps therein.
        This should be called before calling any django submodule functions which
        expect apps to be loaded.

        This function is idempotent.
        """
        if sys.path[0] != '':  # pragma: no cover
            sys.path.insert(0, '')
        django.setup()

    @staticmethod
    def get_models_requiring_annotations():
        """
        Determine all local and non-local models via django model introspection.

        Note that non-local models returned may contain 1st party models (authored by
        edX).  This is a compromise in accuracy in order to simplify the generation
        of this list, and also to ease the transition from zero to 100% annotations
        in edX satellite repositories.

        Returns:
            tuple:
                2-tuple where the first item is a set of local models, and the
                second item is a set of non-local models.
        """
        DjangoSearch.setup_django()
        local_models = set()
        non_local_models = set()
        models_requiring_annotations = []
        total_models = 0

        for app in apps.get_app_configs():
            for root_model in app.get_models():
                total_models += 1
                if DjangoSearch.requires_annotations(root_model):
                    if DjangoSearch.is_non_local(root_model):
                        non_local_models.add(root_model)
                    else:
                        local_models.add(root_model)

                    models_requiring_annotations.append(DjangoSearch.get_model_id(root_model))

        return local_models, non_local_models, total_models, models_requiring_annotations



"""
Base utilities for building annotation-based Sphinx extensions.
"""


from code_annotations.base import AnnotationConfig
from code_annotations.find_static import StaticSearch


def find_annotations(source_path, config_path, group_by_key):
    """
    Find the feature toggles as defined in the configuration file.

    Return:
        toggles (dict): feature toggles indexed by name.
    """
    config = AnnotationConfig(
        config_path, verbosity=-1, source_path_override=source_path
    )
    search = StaticSearch(config)
    all_results = search.search()
    toggles = {}
    for filename in all_results:
        for annotations in search.iter_groups(all_results[filename]):
            current_entry = {}
            for annotation in annotations:
                key = annotation["annotation_token"]
                value = annotation["annotation_data"]
                if key == group_by_key:
                    toggle_name = value
                    toggles[toggle_name] = current_entry
                    current_entry["filename"] = filename
                    current_entry["line_number"] = annotation["line_number"]
                else:
                    current_entry[key] = value

    return toggles


def quote_value(value):
    """
    Quote a Python object if it is string-like.
    """
    if value in ("True", "False", "None"):
        return str(value)
    try:
        float(value)
        return str(value)
    except ValueError:
        pass
    if isinstance(value, str):
        return f'"{value}"'
    return str(value)



"""
Contains functionality for turning YAML reports into human-readable documentation.
"""

import collections
import datetime
import os

import jinja2
import yaml
from slugify import slugify


class ReportRenderer:
    """
    Generates human readable documentation from YAML reports.
    """

    def __init__(self, config, report_files):
        """
        Initialize a ReportRenderer.

        Args:
            config: An AnnotationConfig object
            report_files: A list of files to combine and report on
        """
        self.config = config
        self.echo = self.config.echo
        self.report_files = report_files
        self.create_time = datetime.datetime.now().isoformat()

        self.full_report = self._aggregate_reports()

        self.jinja_environment = jinja2.Environment(
            autoescape=False,
            loader=jinja2.FileSystemLoader(self.config.report_template_dir),
            lstrip_blocks=True,
            trim_blocks=True
        )
        self.top_level_template = self.jinja_environment.get_template('annotation_list.tpl')
        self.all_choices = []
        self.group_mapping = {}

        for token in self.config.choices:
            self.all_choices.extend(self.config.choices[token])

        for group_name in self.config.groups:
            for token in self.config.groups[group_name]:
                self.group_mapping[token] = group_name

    def _add_report_file_to_full_report(self, report_file, report):
        """
        Add a specified report file to a report.

        Args:
            report_file:
            report:

        Returns:

        """
        loaded_report = yaml.safe_load(report_file)

        for filename in loaded_report:
            if filename in report:
                for loaded_annotation in loaded_report[filename]:
                    found = False
                    for report_annotation in report[filename]:
                        index_keys = ('line_number', 'annotation_token', 'annotation_data')

                        if all([loaded_annotation[k] == report_annotation[k] for k in index_keys]):
                            report_annotation.update(loaded_annotation)
                            found = True
                            break

                    if not found:
                        report[filename].append(loaded_annotation)
            else:
                report[filename] = loaded_report[filename]

    def _aggregate_reports(self):
        """
        Combine all of the given report files into a single report object.
        """
        report = collections.defaultdict(list)

        # Combine report files into a single dict. If there are duplicate annotations, make sure we have the superset
        # of data.
        for r in self.report_files:
            self._add_report_file_to_full_report(r, report)

        return report

    def _write_doc_file(self, doc_filename, doc_data):
        """
        Write out a single report file with the given data. This is rendered using the configured top level template.

        Args:
            doc_filename: Filename to write to.
            doc_data: Dict of reporting data to use, in the {'file name': [list, of, annotations,]} style.
        """
        full_doc_filename = os.path.join(
            self.config.rendered_report_dir,
            slugify(doc_filename)
        )

        full_doc_filename += self.config.rendered_report_file_extension

        self.echo.echo_v(f'Writing {full_doc_filename}')

        with open(full_doc_filename, 'w') as output:
            output.write(self.top_level_template.render(
                create_time=self.create_time,
                report=doc_data,
                all_choices=self.all_choices,
                all_annotations=self.config.annotation_tokens,
                group_mapping=self.group_mapping,
                slugify=slugify,
                source_link_prefix=self.config.rendered_report_source_link_prefix)
            )

    def _generate_per_choice_docs(self):
        """
        Generate a page of documentation for each configured annotation choice.
        """
        for choice in self.all_choices:
            choice_report = collections.defaultdict(list)
            for filename in self.full_report:
                for annotation in self.full_report[filename]:
                    if isinstance(annotation['annotation_data'], list) and choice in annotation['annotation_data']:
                        choice_report[filename].append(annotation)

            self._write_doc_file(f'choice_{choice}', choice_report)

    def _generate_per_annotation_docs(self):
        """
        Generate a page of documentation for each configured annotation.
        """
        for annotation in self.config.annotation_tokens:
            annotation_report = collections.defaultdict(list)
            for filename in self.full_report:
                for report_annotation in self.full_report[filename]:
                    if report_annotation['annotation_token'] == annotation:
                        annotation_report[filename].append(report_annotation)

            self._write_doc_file(f'annotation_{annotation}', annotation_report)

    def render(self):
        """
        Perform the rendering of all documentation using the configured Jinja2 templates.
        """
        # Generate the top level list of all annotations
        self._write_doc_file('index', self.full_report)
        self._generate_per_choice_docs()
        self._generate_per_annotation_docs()



"""
Command line interface for code annotation tools.
"""
import datetime
import sys
import traceback

import click

from code_annotations.base import AnnotationConfig, ConfigurationException
from code_annotations.find_django import DjangoSearch
from code_annotations.find_static import StaticSearch
from code_annotations.generate_docs import ReportRenderer
from code_annotations.helpers import fail


@click.group()
def entry_point():
    """
    Top level click command for the code annotation tools.
    """


@entry_point.command('django_find_annotations')
@click.option(
    '--config_file',
    default='.annotations',
    help='Path to the configuration file',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--seed_safelist/--no_safelist',
    default=False,
    show_default=True,
    help='Generate an initial safelist file based on the current Django environment.',
)
@click.option(
    '--list_local_models/--no_list_models',
    default=False,
    show_default=True,
    help='List all locally defined models (in the current repo) that require annotations.',
)
@click.option('--app_name', default='', help='(Optional) App name for which coverage is generated.')
@click.option('--report_path', default=None, help='Location to write the report')
@click.option('-v', '--verbosity', count=True, help='Verbosity level (-v through -vvv)')
@click.option('--lint/--no_lint', help='Enable or disable linting checks', default=False, show_default=True)
@click.option('--report/--no_report', help='Enable or disable writing the report', default=False, show_default=True)
@click.option('--coverage/--no_coverage', help='Enable or disable coverage checks', default=False, show_default=True)
def django_find_annotations(
        config_file,
        seed_safelist,
        list_local_models,
        app_name,
        report_path,
        verbosity,
        lint,
        report,
        coverage
):
    """
    Subcommand for dealing with annotations in Django models.
    """
    try:
        start_time = datetime.datetime.now()
        config = AnnotationConfig(config_file, report_path, verbosity)
        searcher = DjangoSearch(config)

        # Early out if we're trying to do coverage, but a coverage target is not configured
        if coverage and not config.coverage_target:
            raise ConfigurationException("Please add 'coverage_target' to your configuration before running --coverage")

        if seed_safelist:
            searcher.seed_safelist()

        if list_local_models:
            searcher.list_local_models()

        if lint or report or coverage:
            annotated_models = searcher.search()

            if lint:
                click.echo("Performing linting checks...")

                # Check grouping and choices
                if not searcher.check_results(annotated_models):
                    click.secho("\nSearch failed due to linting errors!", fg="red")
                    click.secho("{} errors:".format(len(searcher.errors)), fg="red")
                    click.secho("---------------------------------", fg="red")
                    click.echo("\n".join(searcher.errors))
                    # If there are any errors, do not continue
                    sys.exit(1)
                click.echo("Linting passed without errors.")

            if coverage:
                if not searcher.check_coverage():
                    # If there are any errors, do not continue
                    sys.exit(1)

                click.echo("Coverage passed without errors.")

            if report:
                searcher.report(annotated_models, app_name)

            annotation_count = 0

            for filename in annotated_models:
                annotation_count += len(annotated_models[filename])

            elapsed = datetime.datetime.now() - start_time
            click.echo("Search found {} annotations in {} seconds.".format(
                annotation_count, elapsed.total_seconds()
            ))

    except Exception as exc:  # pylint: disable=broad-except
        click.echo(traceback.print_exc())
        fail(str(exc))


@entry_point.command('static_find_annotations')
@click.option(
    '--config_file',
    default='.annotations',
    help='Path to the configuration file',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--source_path',
    help='Location of the source code to search',
    type=click.Path(exists=True, dir_okay=True, resolve_path=True)
)
@click.option('--report_path', default=None, help='Location to write the report')
@click.option('-v', '--verbosity', count=True, help='Verbosity level (-v through -vvv)')
@click.option('--lint/--no_lint', help='Enable or disable linting checks', default=True, show_default=True)
@click.option('--report/--no_report', help='Enable or disable writing the report file', default=True, show_default=True)
def static_find_annotations(config_file, source_path, report_path, verbosity, lint, report):
    """
    Subcommand to find annotations via static file analysis.
    """
    try:
        start_time = datetime.datetime.now()
        config = AnnotationConfig(config_file, report_path, verbosity, source_path)
        searcher = StaticSearch(config)
        all_results = searcher.search()

        if lint:
            click.echo("Performing linting checks...")
            # Check grouping and choices
            searcher.check_results(all_results)

            # If there are any errors, do not generate the report
            if searcher.errors:
                click.secho("\nSearch failed due to linting errors!", fg="red")
                click.secho("{} errors:".format(len(searcher.errors)), fg="red")
                click.secho("---------------------------------", fg="red")
                click.echo("\n".join(searcher.errors))
                sys.exit(1)
            click.echo("Linting passed without errors.")

        if report:
            click.echo("Writing report...")
            report_filename = searcher.report(all_results)
            click.echo(f"Report written to {report_filename}.")

        elapsed = datetime.datetime.now() - start_time
        annotation_count = 0

        for filename in all_results:
            annotation_count += len(all_results[filename])

        click.echo(f"Search found {annotation_count} annotations in {elapsed}.")

    except Exception as exc:  # pylint: disable=broad-except
        click.echo(traceback.print_exc())
        fail(str(exc))


@entry_point.command("generate_docs")
@click.option(
    '--config_file',
    default='.annotations',
    help='Path to the configuration file',
    type=click.Path(exists=True, dir_okay=False)
)
@click.option('-v', '--verbosity', count=True, help='Verbosity level (-v through -vvv)')
@click.argument("report_files", type=click.File('r'), nargs=-1)
def generate_docs(
        config_file,
        verbosity,
        report_files
):
    """
    Generate documentation from a code annotations report.
    """
    start_time = datetime.datetime.now()

    try:
        config = AnnotationConfig(config_file, verbosity)

        for key in (
                'report_template_dir',
                'rendered_report_dir',
                'rendered_report_file_extension',
                'rendered_report_source_link_prefix'
        ):
            if not getattr(config, key):
                raise ConfigurationException(f"No {key} key in {config_file}")

        config.echo("Rendering the following reports: \n{}".format("\n".join([r.name for r in report_files])))

        renderer = ReportRenderer(config, report_files)
        renderer.render()

        elapsed = datetime.datetime.now() - start_time
        click.echo(f"Report rendered in {elapsed.total_seconds()} seconds.")
    except Exception as exc:  # pylint: disable=broad-except
        click.echo(traceback.print_exc())
        fail(str(exc))



"""
Sphinx extension for viewing (non-toggle) setting annotations.
"""
import os

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective

from code_annotations.contrib.config import SETTING_ANNOTATIONS_CONFIG_PATH

from .base import find_annotations, quote_value


def find_settings(source_path):
    """
    Find the Django settings as defined in the configuration file.

    Return:
        settings (dict): Django settings indexed by name.
    """
    return find_annotations(source_path, SETTING_ANNOTATIONS_CONFIG_PATH, ".. setting_name:")


class Settings(SphinxDirective):
    """
    Sphinx directive to document Django settings in a single documentation page.

    Use this directive as follows::

        .. settings::
            :folder_path: lms/envs/common.py

    This directive supports the following configuration parameters:

    - ``settings_source_path``: absolute path to the repository file tree. E.g:

        settings_source_path = os.path.join(os.path.dirname(__file__), "..", "..")

    - ``settings_repo_url``: Github repository where the code is hosted. E.g:

        settings_repo_url = "https://github.com/edx/myrepo"

    - ``settings_repo_version``: current version of the git repository. E.g:

        import git
        try:
            repo = git.Repo(search_parent_directories=True)
            settings_repo_version = repo.head.object.hexsha
        except git.InvalidGitRepositoryError:
            settings_repo_version = "master"
    """

    required_arguments = 0
    optional_arguments = 1
    option_spec = {"folder_path": directives.unchanged}

    def run(self):
        """
        Public interface of the Directive class.

        Return:
            nodes (list): nodes to be appended to the resulting document.
        """
        return list(self.iter_nodes())

    def iter_nodes(self):
        """
        Iterate on the docutils nodes generated by this directive.
        """
        folder_path = self.options.get("folder_path", "")
        source_path = os.path.join(self.env.config.settings_source_path, folder_path)
        settings = find_settings(source_path)
        # folder_path can point to a file or directory
        root_folder = folder_path if os.path.isdir(source_path) else os.path.dirname(folder_path)
        for setting_name in sorted(settings):
            setting = settings[setting_name]
            # setting["filename"] is relative to the root_path
            setting_filename = os.path.join(root_folder, setting["filename"])
            setting_default_value = setting.get(".. setting_default:", "Not defined")
            setting_default_node = nodes.literal(
                text=quote_value(setting_default_value)
            )
            setting_section = nodes.section("", ids=[f"setting-{setting_name}"])
            setting_section += nodes.title(text=setting_name)
            setting_section += nodes.paragraph("", "Default: ", setting_default_node)
            setting_section += nodes.paragraph(
                "",
                "Source: ",
                nodes.reference(
                    text="{} (line {})".format(
                        setting["filename"], setting["line_number"]
                    ),
                    refuri="{}/blob/{}/{}#L{}".format(
                        self.env.config.settings_repo_url,
                        self.env.config.settings_repo_version,
                        setting_filename,
                        setting["line_number"],
                    ),
                ),
            )
            setting_section += nodes.paragraph(
                text=setting.get(".. setting_description:", "")
            )
            if setting.get(".. setting_warning:") not in (None, "None", "n/a", "N/A"):
                setting_section += nodes.warning(
                    "", nodes.paragraph("", setting[".. setting_warning:"])
                )
            yield setting_section


def setup(app):
    """
    Declare the Sphinx extension.
    """
    app.add_config_value(
        "settings_source_path", os.path.abspath(".."), "env",
    )
    app.add_config_value("settings_repo_url", "", "env")
    app.add_config_value("settings_repo_version", "master", "env")
    app.add_directive("settings", Settings)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
Expose contrib configuration file paths as Python variables, for use in 3rd-party utilities.
"""
import os

import pkg_resources

FEATURE_TOGGLE_ANNOTATIONS_CONFIG_PATH = pkg_resources.resource_filename(
    "code_annotations",
    os.path.join("contrib", "config", "feature_toggle_annotations.yaml"),
)
SETTING_ANNOTATIONS_CONFIG_PATH = pkg_resources.resource_filename(
    "code_annotations",
    os.path.join("contrib", "config", "setting_annotations.yaml"),
)



"""
Sphinx extension for viewing feature toggle annotations.
"""
import os

from docutils import nodes
from sphinx.util.docutils import SphinxDirective

from code_annotations.contrib.config import FEATURE_TOGGLE_ANNOTATIONS_CONFIG_PATH

from .base import find_annotations, quote_value


def find_feature_toggles(source_path):
    """
    Find the feature toggles as defined in the configuration file.

    Return:
        toggles (dict): feature toggles indexed by name.
    """
    return find_annotations(
        source_path, FEATURE_TOGGLE_ANNOTATIONS_CONFIG_PATH, ".. toggle_name:"
    )


class FeatureToggles(SphinxDirective):
    """
    Sphinx directive to list the feature toggles in a single documentation page.

    Use this directive as follows::

        .. featuretoggles::

    This directive supports the following configuration parameters:

    - ``featuretoggles_source_path``: absolute path to the repository file tree. E.g:

        featuretoggles_source_path = os.path.join(os.path.dirname(__file__), "..", "..")

    - ``featuretoggles_repo_url``: Github repository where the code is hosted. E.g:

        featuretoggles_repo_url = "https://github.com/edx/myrepo"

    - ``featuretoggles_repo_version``: current version of the git repository. E.g:

        import git
        try:
            repo = git.Repo(search_parent_directories=True)
            featuretoggles_repo_version = repo.head.object.hexsha
        except git.InvalidGitRepositoryError:
            featuretoggles_repo_version = "master"
    """

    required_arguments = 0
    optional_arguments = 0
    option_spec = {}

    def run(self):
        """
        Public interface of the Directive class.

        Return:
            nodes (list): nodes to be appended to the resulting document.
        """
        return list(self.iter_nodes())

    def iter_nodes(self):
        """
        Iterate on the docutils nodes generated by this directive.
        """
        toggles = find_feature_toggles(self.env.config.featuretoggles_source_path)
        for toggle_name in sorted(toggles):
            toggle = toggles[toggle_name]
            toggle_default_value = toggle.get(".. toggle_default:", "Not defined")
            toggle_default_node = nodes.literal(text=quote_value(toggle_default_value))
            toggle_section = nodes.section("", ids=[f"featuretoggle-{toggle_name}"])
            toggle_section += nodes.title(text=toggle_name)
            toggle_section += nodes.paragraph("", "Default: ", toggle_default_node)
            toggle_section += nodes.paragraph(
                "",
                "Source: ",
                nodes.reference(
                    text="{} (line {})".format(
                        toggle["filename"], toggle["line_number"]
                    ),
                    refuri="{}/blob/{}/{}#L{}".format(
                        self.env.config.featuretoggles_repo_url,
                        self.env.config.featuretoggles_repo_version,
                        toggle["filename"],
                        toggle["line_number"],
                    ),
                ),
            )
            toggle_section += nodes.paragraph(
                text=toggle.get(".. toggle_description:", "")
            )
            if toggle.get(".. toggle_warnings:") not in (None, "None", "n/a", "N/A"):
                toggle_section += nodes.warning(
                    "", nodes.paragraph("", toggle[".. toggle_warnings:"])
                )
            yield toggle_section


def setup(app):
    """
    Declare the Sphinx extension.
    """
    app.add_config_value(
        "featuretoggles_source_path",
        os.path.abspath(".."),
        "env",
    )
    app.add_config_value("featuretoggles_repo_url", "", "env")
    app.add_config_value("featuretoggles_repo_version", "master", "env")
    app.add_directive("featuretoggles", FeatureToggles)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
Helpers for code_annotations scripts.
"""
import os
import re
import sys
from io import StringIO
from pprint import pprint

import click


def fail(msg):
    """
    Log the message and exit.

    Args:
        msg: Message to log
    """
    click.secho(msg, fg="red")
    sys.exit(1)


class VerboseEcho:
    """
    Helper to handle verbosity-dependent logging.
    """

    verbosity = 1

    def __call__(self, output, **kwargs):
        """
        Echo the given output regardless of verbosity level.

        This is just a convenience method to avoid lots of `self.echo.echo()`.

        Args:
            output: Text to output
            kwargs: Any additional keyword args to pass to click.echo
        """
        self.echo(output, **kwargs)

    def set_verbosity(self, verbosity):
        """
        Override the default verbosity level.

        Args:
            verbosity: The verbosity level to set to
            kwargs: Any additional keyword args to pass to click.echo
        """
        self.verbosity = verbosity
        self.echo_v(f"Verbosity level set to {verbosity}")

    def echo(self, output, verbosity_level=0, **kwargs):
        """
        Echo the given output, if over the verbosity threshold.

        Args:
            output: Text to output
            verbosity_level: Only output if our verbosity level is >= this.
            kwargs: Any additional keyword args to pass to click.echo
        """
        if verbosity_level <= self.verbosity:
            click.secho(output, **kwargs)

    def echo_v(self, output, **kwargs):
        """
        Echo the given output if verbosity level is >= 1.

        Args:
            output: Text to output
            kwargs: Any additional keyword args to pass to click.echo
        """
        self.echo(output, 1, **kwargs)

    def echo_vv(self, output, **kwargs):
        """
        Echo the given output if verbosity level is >= 2.

        Args:
            output: Text to output
            kwargs: Any additional keyword args to pass to click.echo
        """
        self.echo(output, 2, **kwargs)

    def echo_vvv(self, output, **kwargs):
        """
        Echo the given output if verbosity level is >= 3.

        Args:
            output: Text to output
            kwargs: Any additional keyword args to pass to click.echo
        """
        self.echo(output, 3, **kwargs)

    def pprint(self, data, indent=4, verbosity_level=0):
        """
        Pretty-print some data with the given verbosity level.
        """
        formatted = StringIO()
        pprint(data, indent=indent, stream=formatted)
        formatted.seek(0)
        self.echo(formatted.read(), verbosity_level=verbosity_level)


def clean_abs_path(filename_to_clean, parent_path):
    """
    Safely strips the parent path from the given filename, leaving only the relative path.

    Args:
        filename_to_clean: Input filename
        parent_path: Path to remove from the input

    Returns:
        Updated path
    """
    # If we are operating on only one file we don't know what to strip off here,
    # just return the whole thing.
    if filename_to_clean == parent_path:
        return os.path.basename(filename_to_clean)
    return os.path.relpath(filename_to_clean, parent_path)


def get_annotation_regex(annotation_regexes):
    """
    Return the full regex to search inside comments for configured annotations.

    A successful match against the regex will return two groups of interest: 'token'
    and 'data'.

    This regular expression supports annotation tokens that span multiple lines. To do
    so, prefix each line after the first by at least two leading spaces. E.g:

        .. pii: First line
          second line

    Unfortunately, the indenting spaces will find their way to the content of the "token" group.

    Args:
        annotation_regexes: List of re.escaped annotation tokens to search for.

    Returns:
        Regex ready for searching comments for annotations.
    """
    annotation_regex = r"""
    (?P<space>[\ \t]*)               # Leading empty spaces
    (?P<token>{tokens})              # Python format string that will be replaced with a
                                     # regex, escaped and then or-joined to make a list
                                     # of the annotation tokens we're looking for
                                     # Ex: (\.\.\ pii\:\:|\.\.\ pii\_types\:\:)
    (?P<data>                        # Captured annotation data
        (?:                          # non-capture mode
            .                        # any non-newline character
            |                        # or new line of multi-line annotation data
            (?:                      # non-capture mode
                \n{{1,}}             # at least one newline,
                (?P=space)           # followed by as much space as the prefix,
                (?P<indent>\ {{2,}}) # at least two spaces,
                (?=[^\ ])            # and a non-space character (look-ahead)
                (?!{tokens})         # that does not match any of the token regexes
            )                        #
        )*                           # any number of times
    )
    """
    annotation_regex = annotation_regex.format(tokens='|'.join(annotation_regexes))
    return re.compile(annotation_regex, flags=re.VERBOSE)


def clean_annotation(token, data):
    """
    Clean annotation token and data by stripping all trailing/prefix empty spaces.

    Args:
        token (str)
        data (str)

    Returns:
        (str, str): Tuple of cleaned token, data
    """
    token = token.strip()
    data = data.strip()
    return token, data



"""
List possible annotation error types.
"""
from collections import namedtuple

AnnotationErrorType = namedtuple(
    "AnnotationError", ["message", "symbol", "description"]
)

# The TYPES list should contain all AnnotationErrorType instances. This list can then be parsed by others, for instance
# to expose errors to pylint.
TYPES = []


def add_error_type(message, symbol, description):
    """
    Create an AnnotationErrorType instance and add it to TYPES.
    """
    error_type = AnnotationErrorType(
        message,
        symbol,
        description,
    )
    TYPES.append(error_type)
    if len(TYPES) > 10:
        # if more than 10 items are created here, numerical IDs generated by edx-lint will overlap with other warning
        # IDs.
        raise ValueError("TYPES may not contain more than 10 items")
    return error_type


# It is important to preserve the insertion order of these error types in the TYPES list, as edx-lint uses the error
# type indices to generate numerical pylint IDs. If the insertion order is changed, the pylint IDs will change too,
# which might cause incompatibilities down the road. Thus, new items should be added at the end.
InvalidChoice = add_error_type(
    '"%s" is not a valid choice for "%s". Expected one of %s.',
    "annotation-invalid-choice",
    "Emitted when the value of a choice field is not one of the valid choices",
)
DuplicateChoiceValue = add_error_type(
    '"%s" is already present in this annotation.',
    "annotation-duplicate-choice-value",
    "Emitted when duplicate values are found in a choice field",
)
MissingChoiceValue = add_error_type(
    'no value found for "%s". Expected one of %s.',
    "annotation-missing-choice-value",
    "Emitted when a choice field does not have any value",
)
InvalidToken = add_error_type(
    "'%s' token does not belong to group '%s'. Expected one of: %s",
    "annotation-invalid-token",
    "Emitted when a token is found in a group for which it is not valid",
)
DuplicateToken = add_error_type(
    "found duplicate token '%s'",
    "annotation-duplicate-token",
    "Emitted when a token is found twice in a group",
)
MissingToken = add_error_type(
    "missing non-optional annotation: '%s'",
    "annotation-missing-token",
    "Emitted when a required token is missing from a group",
)
